#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""PostgreSQL in Thread v0.07
From iamvic on www.cyberforum.ru
Restrictions: Python 3.x, PyQt and no more.
"""

import copy

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import QMutex, QMutexLocker
from PyQt5.QtCore import QObject, QThread
from PyQt5.QtWidgets import QApplication
from PyQt5.QtSql import QSqlDatabase, QSqlQuery

class Thread_postgresql(QThread):
    connect_parm = [
        'localhost',    # хост
        '5432',         # порт
        'postgres',     # база данных
        'postgres',     # пользователь
        'postgres'      # пароль
        ]
    thread_mutex = QMutex()
    connection_created = pyqtSignal()
    duplicated_thread = pyqtSignal()
    invalid_using = pyqtSignal()

    def set_connection_parm(self, parm):
        '''передать потоку параметры соединения извне
        (для наружного употребления)
        '''
        with QMutexLocker(self.thread_mutex):
            rc = (type(parm) == list and
                  len(parm) == 5 and
                  all(type(x) == str for x in parm) and
                  parm[1].isdigit() and
                  int(parm[1]) in range(1,65535))
            if rc:
                self.connect_parm = copy.deepcopy(parm)
        return rc

    def get_connection_parm(self):
        '''получить переданные снаружи параметры соединения
        (для внутреннего употребления)
        '''
        with QMutexLocker(self.thread_mutex):
            rc = copy.deepcopy(self.connect_parm)
        return rc

    def run(self):
        '''именно этот код будет выполняться в дополнительном потоке
        после вызова QThread.start()
        '''
        def inner_run(connect_name, current_parm):
            '''поскольку удалить соединение можно только вне области
            видимости соединения, то его создание и вся работа с БД
            выполняется только в теле этой функции, а удаление соединения
            производится снаружи после выхода из неё
            '''
            db = QSqlDatabase.addDatabase('QPSQL', connect_name)
            db.setHostName(current_parm[0])
            db.setPort(int(current_parm[1]))
            db.setDatabaseName(current_parm[2])
            db.setUserName(current_parm[3])
            db.setPassword(current_parm[4])

            if not db.open():
                # сигнал об ошибке с информацией из db.lastError().text()
                pass
            else:
                self.connection_created.emit()
                query = QSqlQuery(db)
                if query.exec(
                    "SET application_name TO '{!s}';".format(connect_name)):
                    # хоть какая-то идентификация для админа БД.
                    # QSqlQuery.prepare() с SET не работает
                    # (видимо, о нём не знает).
                    pass
                else:
                    # сигнал об ошибке с информацией из query.lastError().text()
                    pass

                for i in range(10):
                    QThread.sleep(1)
                    if self.isInterruptionRequested():
                        # если запрошено прерывание,
                        # то вываливаемся из цикла и завершаем работу
                        break
                # этот цикл моделирует работу с БД, она может сопровождаться
                # периодической отправкой сигналов (если надо),
                # а мы просто спим 10 сек, периодически проверяя
                # наличие запроса на прерывание.
                # self.sleep() не используем, потому как self живёт
                # в основном потоке (мало ли что приключится).
                # в качестве местной экзотики :) можно было бы использовать
                # proxy.thread().sleep(), поскольку вспомогательный
                # объект живёт как раз в дополнительном потоке.

            db.close()
            # закрыли БД
            return
            # и вышли из функции

        proxy = QObject()
        # именовать соединение будем по идентификатору доп.потока.
        # чтобы убедиться в том, что для запуска использовался именно
        # метод QThread.start(), создаём вспомогательный объект,
        # живущий в дополнительном потоке (self живёт в основном).
        # вспомогательный объект не имеет родителя, поскольку
        # он - первый и, в данном случае, единственный на весь доп.поток.
        # а родитель, как известно, должен жить в том же потоке,
        # что и его отпрыски.

        if id(self) == id(proxy.thread()):
            # вспомогательный объект действительно живёт
            # в дополнительном потоке (т.е. в self)
            # => использовали именно QThread.start()

            connect_name = 'thread_{!s}_main_{!s}_({!s})'.format(
                # по идентификаторам дополнительного и основного потоков
                # создали уникальное имя соединения вида
                # thread_0x12345678_main_0x1234abcd_(True)
                hex(id(self)),
                # thread_0x12345678_...
                hex(id(self.thread())),
                # ...main_0x1234abcd_...
                (id(QApplication.instance().thread()) ==
                 id(self.thread()) == id(proxy.thread().thread()) )
                # ...(True)
                # (True) подтверждает, что дополнительным потоком
                # управляет именно основной (вот такая паранойя :)
                )

            if any(s.split('_')[1] == hex(id(proxy.thread()))
                   for s in QSqlDatabase.connectionNames()):
                # параноидальная проверка с поиском дубликата
                # идентификатора дополнительного потока
                # в именах существующих соединений
                # (совершенно лишняя, но пусть будет)
                # дубликат найден => сообщаем и выходим
                self.duplicated_thread.emit()

            else:
                # дубликат не найден => продолжаем работу
                current_parm = self.get_connection_parm()
                # запросили параметры соединения, переданные снаружи

                inner_run(connect_name, current_parm)
                # отработали с базой данных

                QSqlDatabase.removeDatabase(connect_name)
                # удалили соединение

        else:
            # вспомогательный объект живёт в потоке,
            # отличном от дополнительного (т.е. не в self)
            # => использовали QThread.run() вместо QThread.start()
            self.invalid_using.emit()
            # отправляем сигнал о неправильном использовании

        pass
        # завершили выполнение потока
        # если дополнительный поток был запущен методом QThread.start(),
        # то это вызовет отправку сигнала QThread.finished()
